// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/ADADIO/ADADIO_Linux_3.x.x.x_GSC_DN/driver/device.c $
// $Rev: 52446 $
// $Date: 2023-01-30 13:41:09 -0600 (Mon, 30 Jan 2023) $

// ADADIO: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START		D28		// BCR
#define	INIT_START			D31		// BCR



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model	Vendor	Device	SubVen	SubDev	type
	// Identification for the ADADIO2 model is checked later as the id values
	// are the same as for the ADADIO.
	{ "ADADIO",	0x10B5, 0x9080, 0x10B5, 0x2370,	GSC_DEV_TYPE_ADADIO	},
	// The old versions of the ADADIO used the 9060.
	{ "ADADIO",	0x10B5, 0x9060, 0x10B5, 0x2370,	GSC_DEV_TYPE_ADADIO	},
	{ NULL }
};



//*****************************************************************************
static int _brr_compute(dev_data_t* dev)
{
	u32		bcr;
	u32		brr;
	long	ms;
	int		ret	= 0;

	// We don't call these IOCTL service directly as neither will be done given
	// that interrupts are not setup at this point. Thus, we do both manually.

	for (;;)	// A convenience loop.
	{
		// Prepare the device for use.
		ret	= initialize_ioctl(dev, NULL);
		if (ret)	break;

		// Initiate auto-calibration.
		os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, AUTO_CAL_START);

		// Wait for completion.
		ms	= dev->cache.auto_cal_ms + 1000;
		os_time_sleep_ms(ms);

		// Check for completion.
		bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

		if (bcr & AUTO_CAL_START)
		{
			ret	= -EIO;
			printk(	"%s: AUTO-CALIBRATION DID NOT COMPLETE (%ld ms)\n",
					dev->model,
					(long) ms);
			break;
		}

		brr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_srr_32);	// same location
		dev->cache.brr	= brr;

		// Invoke Initialization.
		os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, INIT_START, INIT_START);

		// Wait for completion.
		ms	= dev->cache.initialize_ms + 1000;
		os_time_sleep_ms(ms);

		if (bcr & INIT_START)
		{
			ret	= -EIO;
			printk(	"%s: INITIALIZATION DID NOT COMPLETE (%ld ms)\n",
					dev->model,
					(long) ms);
			break;
		}

		break;
	}

	return(ret);
}



//*****************************************************************************
static void _dmdma_compute(dev_data_t* dev)
{
	if (dev->board_type == GSC_DEV_TYPE_ADADIO2)
		dev->cache.dmdma	= 1;
	else
		dev->cache.dmdma	= ((dev->cache.brr & 0x18000) == 0x18000) ? 1 : 0;
}



//*****************************************************************************
static void _fsamp_compute(dev_data_t* dev)
{
	s32	range	= GSC_FIELD_DECODE(dev->cache.brr, 16, 16);

	if (range)
	{
		// 200K S/S
		dev->cache.fsamp_max			= 200000L;
		dev->cache.rate_gen_nrate_min	= 100;
	}
	else
	{
		// 100K S/S
		dev->cache.fsamp_max			= 100000L;
		dev->cache.rate_gen_nrate_min	= 200;
	}

	dev->cache.fsamp_min			= 305L;
	dev->cache.rate_gen_nrate_mask	= 0xFFFF;
	dev->cache.rate_gen_nrate_max	= 0xFFFF;
}



//*****************************************************************************
static void _model_compute(dev_data_t* dev)
{
	u32	adadio2;

	adadio2	= dev->cache.brr & D13;

	if (adadio2)
	{
		dev->board_type	= GSC_DEV_TYPE_ADADIO2;
		dev->model		= "ADADIO2";
	}
}



//*****************************************************************************
static void _vrange_compute(dev_data_t* dev)
{
	s32	range	= GSC_FIELD_DECODE(dev->cache.brr, 18, 17);

	switch (range)
	{
		default:
		case 0:	dev->cache.vrange	= ADADIO_VRANGE_10;		break;
		case 1:	dev->cache.vrange	= ADADIO_VRANGE_5;		break;
		case 2:	dev->cache.vrange	= ADADIO_VRANGE_2_5;	break;
	}
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(ADADIO_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_aidr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_AIDR));
		dev->vaddr.gsc_aoc0r_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_AOC0R));
		dev->vaddr.gsc_aoc1r_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_AOC1R));
		dev->vaddr.gsc_aoc2r_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_AOC2R));
		dev->vaddr.gsc_aoc3r_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_AOC3R));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_BCR));
		dev->vaddr.gsc_diopr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_DIOPR));
		dev->vaddr.gsc_srr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(ADADIO_GSC_SRR));

		// Data cache initialization.
		dev->cache.auto_cal_ms			= 10000;
		dev->cache.fifo_size			= _32K;
		dev->cache.initialize_ms		= 3;
		dev->cache.master_clock			= 20000000L;

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_READ
			| GSC_DMA_CAP_DMDMA_READ;
		ret	= gsc_dma_create(dev, dma, dma);
		if (ret)	break;

		ret	= _brr_compute(dev);
		_model_compute(dev);
		_dmdma_compute(dev);
		_vrange_compute(dev);
		_fsamp_compute(dev);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


